local BugEater = LibStub:NewLibrary("BugEaterAPI-1.0", 3)

BugEater.Buffer 			 = { ERROR = {}, DEBUG = {} }
BugEater.d 					 = d -- store original d()
CHAT_SYSTEM.OrgAddMessage	 = CHAT_SYSTEM.AddMessage -- store original AddMessage()

local SuppressErrorDialog	 = false
local ToggleSuppressDialog	 = ZO_UIErrors_ToggleSupressDialog
local BugEater_SaveData		 = nil
local BugEater_SettingsPanel = nil
local LMP 					 = LibStub("LibMediaProvider-1.0")
local LAM 					 = LibStub("LibAddonMenu-1.0")

local LibOrangUtils = LibStub("LibOrangUtils-1.0")

-----------------------------
-- FILTER CHECK
-----------------------------

local function BugEater_FilterMatch(text, filter)
	if(not text) then do return false end end
	
	local filter = filter or "DEBUG"
	local filter_content = ""
	
	if(filter == "ERROR" and _G["BugEater_ErrorFilter"] and _G["BugEater_ErrorFilter"]["edit"]) then
		filter_content = _G["BugEater_ErrorFilter"]["edit"]:GetText()
	elseif(_G["BugEater_AddonOutputFilter"] and _G["BugEater_AddonOutputFilter"]["edit"]) then
		filter_content = _G["BugEater_AddonOutputFilter"]["edit"]:GetText()
	end

	for line in string.gmatch(filter_content, "[^\n]+") do
		if(text:match(line)) then do return true end end
	end
	
	return false
end

-----------------------------
-- HOOKS
-----------------------------

-- function d(...)
d = function(...) 
-- Yes we overwrite the original d() because it doesn't properly buffer the output
	for i = 1, select("#", ...) do
		local msg = select(i, ...)
		
		BugEater:AddMessage(msg, "DEBUG")
		
		if(BugEater_SaveData and BugEater_SaveData.SuppressAddonOutput == true) then do return end end
		
		if(type(msg) == "string") then 
			-- suppress gui warnings
			if(BugEater_SaveData and BugEater_SaveData.SuppressGuiWarnings == true and (msg:find("Gui Warning:") or msg:find("many anchors processed"))) then do return end end	
			
			-- check if any custom filter matched 
			if(BugEater_FilterMatch(msg)) then do return end end	
		end
		
		BugEater.d(msg) 
	end
end

function CHAT_SYSTEM:AddMessage(msg)
	if(BugEater_SaveData and BugEater_SaveData.SuppressAddonOutput == true) then do return end end
		
	-- suppress gui warnings
	if(BugEater_SaveData and BugEater_SaveData.SuppressGuiWarnings == true and (msg:find("Gui Warning:") or msg:find("many anchors processed"))) then do return end end	
		
	-- check if any custom filter matched 
	if(BugEater_FilterMatch(msg)) then do return end end	
		
	CHAT_SYSTEM:OrgAddMessage(msg)
end

-----------------------------
-- API
-----------------------------

function BugEater:AddMessage(msg, buffer)
	if(not msg or not buffer or	BugEater.Buffer[buffer] == nil) then do return "" end end

	local buf = BugEater.Buffer[buffer]
	
	buf[#buf + 1] = msg
	
	return buf[#buf + 1]
end

function BugEater:GetMessage(index, buffer)
	if(not index or not buffer or BugEater.Buffer[buffer] == nil) then do return "" end end
	
	local msg = BugEater.Buffer[buffer][index] or ""
	
	return msg
end

function BugEater:PrintMessages(buffer)
	if(BugEater.Buffer[buffer] == nil) then do return false end end
	
	for i,msg in ipairs(BugEater.Buffer[buffer]) do
		BugEater.d(msg)
	end
	
	return true
end

function BugEater:GetNumMessages(buffer)
	if(BugEater.Buffer[buffer] == nil) then do return -1 end end
		
	local buf = BugEater.Buffer[buffer]
	
	return #buf
end

function BugEater:AddBuffer(buffer)
	if(not buffer or BugEater.Buffer[buffer] ~= nil) then do return false end end
	
	BugEater.Buffer[buffer] = {}
	
	return true
end

function BugEater:GetBuffer(buffer)
	if(not buffer or BugEater.Buffer[buffer] == nil) then do return false end end
	
	return BugEater.Buffer[buffer]
end

function BugEater:FlushBuffer(buffer)
	if(BugEater.Buffer[buffer] ~= nil) then
		BugEater.Buffer[buffer] = {}
	elseif(buffer == "ALL") then
		for k,v in pairs(BugEater.Buffer) do
			BugEater.Buffer[k] = {}
		end
	else
		return false
	end
end

function BugEater:EnableErrorDialog()
	if(SuppressErrorDialog == true) then		
		ToggleSuppressDialog()
		ZO_UIErrors:SetAlpha(1)
		
		SuppressErrorDialog = not SuppressErrorDialog
	end
end

function BugEater:DisableErrorDialog()
	if(SuppressErrorDialog ~= true) then		
		ToggleSuppressDialog() -- yes this is dirty, but there is no other reliable way to Suppress the error-dialogue in case our lib also throws an error
		ZO_UIErrors:SetAlpha(0) -- just in case toggle didn't work or got changed from another addon
		
		SuppressErrorDialog = not SuppressErrorDialog
	end
end

-----------------------------
-- PRIVATE FUNCTIONS
-----------------------------

-- Check if Error is Unique
local function BugEater_IsErrorUnique(error_thrown)
	for _,error_buffered in ipairs(BugEater.Buffer.ERROR) do
		if(error_buffered == error_thrown) then do return false end end
	end

	return true
end

-- Merging EDIT + TextBuffer = EditBuffer :)
local function BugEater_CreateEditBuffer(name, parent, anchor1, anchor2, fontSize, numLines, lineHeight)
	if(not name or not parent) then do return nil end end

	local anchor1_default = {TOPLEFT, parent, TOPLEFT, 0, 0}
	local anchor2_default = {BOTTOMRIGHT, parent, BOTTOMRIGHT, 0, 0}
	
	-- Optional Parameters
	local anchor1 	= anchor1 or anchor1_default
	local anchor2 	= anchor2 or anchor2_default
	
	if(type(anchor1) ~= "table") then anchor1 = anchor1_default end
	if(type(anchor2) ~= "table") then anchor2 = anchor2_default end
	
	--'BugEater/fonts/AnkaCoder-C87-r.ttf' size:17 charwidth: 8.75 | 21 @ 11.5 | 20 @ 10.75
	-- Font settings
	local FONT_STYLE_THIN = 'soft-shadow-thin'
	local FONT_STYLE_THICK = 'soft-shadow-thick'
	local FONT_STYLE_SHADOW = 'shadow'
	local FONT_STYLE_NONE = 'none'	
	local FONT_SIZE = fontSize or 20
	local FONT_FACE = 'BugEater/fonts/AnkaCoder-C87-b.ttf'
	local CHAR_WIDTH = 10.75
	local LINE_HEIGHT = lineHeight or tonumber(FONT_SIZE*1.275)
	
	-- Control-Names
	local BG = name..'_BG'

	-- BG
	WINDOW_MANAGER:CreateControlFromVirtual(BG, parent, "ZO_InsetBackdrop")
	--_G[BG]:SetAlpha(0.75)
	_G[BG]:SetAnchor(anchor1[1], anchor1[2], anchor1[3], anchor1[4], anchor1[5])	
	_G[BG]:SetAnchor(anchor2[1], anchor2[2], anchor2[3], anchor2[4], anchor2[5])	
	
	-- LINES (EDIT)
	local numLines = numLines or math.floor((_G[BG]:GetHeight()-anchor1[5]-anchor2[5])/LINE_HEIGHT)
	
	for i=1, numLines, 1 do
		local LINE = name..'_LINE'..tostring(i)	
		
		WINDOW_MANAGER:CreateControlFromVirtual(LINE, parent, "ZO_DefaultEditMultiLineForBackdrop")
		_G[LINE]:SetMaxInputChars(1000) -- Hello Windows 98 and FAT, I missed you NOT
		_G[LINE]:SetMultiLine(false)
		_G[LINE]:SetEditEnabled(false)
		_G[LINE]:SetCopyEnabled(true)
		_G[LINE]:SetAnchor(TOPLEFT, _G[BG], TOPLEFT, 7, (5+((i-1)*LINE_HEIGHT)))
		_G[LINE]:SetDimensions((_G[BG]:GetWidth()-10), LINE_HEIGHT)
		_G[LINE]:SetColor(207/255, 220/255, 189/255, 1)
		_G[LINE]:SetHandler("OnMouseDoubleClick", function(self) 
			-- bah
			zo_callLater(function() self:SelectAll() end, 0.5)
		end) 		
		
		-- Set monospace font with fixed width
		_G[LINE]:SetFont(FONT_FACE..'|'..FONT_SIZE..'|'..FONT_STYLE_THIN)
	end
	
	-- Create an empty table / pseudo-control that will hold our public stuff
	local control = {}
	
	control.text 	= "" 	-- holds the unmodified text
	control.offset 	= 1 	-- starts at line 1
	
	-- links & colors
	local function disable_special(text)
		local special = {}
	
		-- links
		for link in string.gfind(text, "(|H(.-):(.-)|h(.-)|h)") do
			local name = link:match("|h(.-)|h")
			if(name) then 
				--BugEater.d("Found Link: "..name)
			
				special[#special+1] = {["name"] = name, ["special"] = link }
				text = LibOrangUtils:str_replace(text, LibOrangUtils:str_escape(link), tostring(name), false) 
			end
		end
		
		-- /dump "|cFFAE19 AAAAAAAA|h "
		-- /dump "|cFFAE19 BBBB |cFFFAAA CCCCC "
		-- /dump "|cFAFAFA DDDD"
		-- colors (|cFFAE19)

		for color in string.gfind(text, "(|c(%x+)([^|]+))") do
			local content = color:match("|c%x+([^|]+)")
			
			special[#special+1] = {["name"] = content, ["special"] = color }
			text = LibOrangUtils:str_replace(text, LibOrangUtils:str_escape(color), tostring(content), false) 
			
			--BugEater.d("FOUND COLOR: "..tostring(content))
		end
		
		return text, special
	end
	
	-- links & colors
	local function enable_special(text, special)
		if(not special or type(special) ~= "table") then do return text end end
	
		for i,o in ipairs(special) do
			local disabled_special = tostring(o["name"])
			
			--BugEater.d("disabled_special: "..disabled_special)
			
			text = LibOrangUtils:str_replace(text, LibOrangUtils:str_escape(disabled_special), o["special"], false) 
		end
		
		return text
	end
	
	local function clear_special(special)
		for i,o in ipairs(special) do
			special[i] = nil
		end
		
		special = nil
		
		return nil
	end
	
	function control:SetText(text) 
		self.text = text
		
		self:Clear()
	
		local numLines = self:GetNumCurLines()
		local maxLines = self:GetNumMaxLines()
		
		if(numLines > maxLines) then
			local offset = numLines-maxLines
						
			self:DisplayLinesAt(offset)
		else
			self:DisplayLinesAt()
		end
		
		if(BugEater_Window_SCROLLBAR) then 
			local max_scroll = math.floor(self:GetNumCurLines()-self:GetNumMaxLines())
			if(max_scroll < 1) then max_scroll = 1 end
			
			BugEater_Window_SCROLLBAR:SetMinMax(1, max_scroll) 
			BugEater_Window_SCROLLBAR:SetValue(max_scroll)
		end
	end
	
	function control:GetNumCurLines() 
		local all_lines = self:GetAllLines()
	
		return #all_lines
	end	
	
	function control:GetNumMaxLines() 
		return numLines
	end		
	
	function control:GetAllLines()
		local ret = {}	
				
		for line in string.gmatch(self.text, "[^\n]+") do
			local line, special = disable_special(line)
			local real_lines = self:SplitLongLines(line)	
						
			for _,r_line in ipairs(real_lines) do
				ret[#ret+1] = enable_special(r_line, special)
			end
			
			special = clear_special(special)
			--all_lines[#all_lines+1] = line
		end	
	
		enable_special()
	
		return ret
	end
	
	function control:DisplayLinesAt(offset)
		self:Clear()
	
		local offset = offset or 0
		
		if(offset < 0) then do return end end
		
		local all_lines = self:GetAllLines()
		
		-- output lines
		for i=1, numLines, 1 do
		
			local LINE = name..'_LINE'..tostring(i)	
		
			if(all_lines[i+offset] and _G[LINE]) then
				_G[LINE]:SetText(all_lines[i+offset])
			end
		end
		
		self.offset = offset
	end
	
	function control:ScrollUp()
		local numLines = self:GetNumCurLines()
		local maxLines = self:GetNumMaxLines()	
		
		-- only scroll if theres more text available then displayed
		if(numLines > maxLines) then
			local new_offset = math.floor(control.offset - 1)
			if(new_offset >= 0) then 
				if(BugEater_Window_SCROLLBAR) then 			
					BugEater_Window_SCROLLBAR:SetValue(new_offset)
				else
					self:DisplayLinesAt(new_offset) 
				end				
			end
		end
	end
	
	function control:ScrollDown()
		local numLines = self:GetNumCurLines()
		local maxLines = self:GetNumMaxLines()	
		
		-- only scroll if theres more text available then displayed
		if(numLines > maxLines) then
			local new_offset = math.floor(control.offset + 1)
			local displayed_lines = new_offset+maxLines
			if(new_offset >= 0 and displayed_lines <= numLines) then 
				if(BugEater_Window_SCROLLBAR) then 			
					BugEater_Window_SCROLLBAR:SetValue(new_offset)
				else
					self:DisplayLinesAt(new_offset) 
				end		
			end		
		end	
	end
	
	function control:ScrollToBottom()
		local numLines = self:GetNumCurLines()
		local maxLines = self:GetNumMaxLines()	
		
		-- only scroll if theres more text available then displayed
		if(numLines > maxLines) then
			local new_offset = math.floor(numLines-maxLines)
			local displayed_lines = new_offset+maxLines
			if(new_offset >= 0 and displayed_lines <= numLines) then 
				if(BugEater_Window_SCROLLBAR) then 			
					BugEater_Window_SCROLLBAR:SetValue(new_offset)
				else
					self:DisplayLinesAt(new_offset) 
				end		
			end		
		end	
	end	
	
	function control:SplitLongLines(text, maxLength)
		local ret = {}
		
		local maxLineWidth 	= _G[BG]:GetWidth()
		local maxChars 		= maxLength or math.floor(maxLineWidth/CHAR_WIDTH)
		local text_len 		= string.len(text);
		
		if(text_len <= maxChars) then
			ret[#ret+1] = text
		else
			for from = 0, text_len, maxChars do
				local to = from+maxChars-1
				
				if(from >= text_len) then from = text_len end
				if(to >= text_len) then to = text_len end
			
				--ret = ret+1
				ret[#ret+1] = string.sub(text, from, to)
			end
		end
		
		return ret
	end	
	
	function control:Clear()
		for i=1, numLines, 1 do
			local LINE = name..'_LINE'..tostring(i)	

			if(_G[LINE]) then _G[LINE]:Clear() end
		end
	end

	-- dirty but butt
	function control:CopyAllTextToClipboard()
		LibOrangUtils:CopyToClipboard(self.text)
	end		
		
	-- OnMouseWheel-Event
	parent:SetHandler("OnMouseWheel", function(self, delta, ctrl, alt, shift, command) 
		-- up
		if(delta > 0) then
			control:ScrollUp()
		-- down
		elseif(delta < 0) then
			control:ScrollDown()
		end
	end)

	_G[name] = control	
	
	return _G[name]

end

local function BugEater_LoadSaves()
	Default	= {
		["SuppressErrorDialog"] = true,
		["ShowErrorsInChat"] = true,
		["AvoidErrorFlood"] = true,
		["ShowPreInitAOP"] = false,
		["SuppressAddonOutput"] = false,
		["SuppressGuiWarnings"] = false,
		["AddonOutputFilter"] = "",
		["ErrorFilter"] = ""
	}

	BugEater_SaveData = ZO_SavedVars:NewAccountWide( "BugEaterSaveData", 1, nil, Default, nil )
end

local function BugEater_CreateSettings()
	LAM:AddHeader(BugEater_SettingsPanel, "BugEater_Settings_General", "BugEater v1.0")
	
	-- Suppress Error Dialog
	LAM:AddCheckbox(BugEater_SettingsPanel, "BugEater_SuppressErrorDialog", "Suppress Error-Dialog", "Suppress the native Error-Dialog?", 
		function() return BugEater_SaveData.SuppressErrorDialog end,	-- getFunc
		function() 														-- setFunc
			BugEater_SaveData.SuppressErrorDialog = not BugEater_SaveData.SuppressErrorDialog
			
			if(BugEater_SaveData.SuppressErrorDialog == true) then
				BugEater.DisableErrorDialog()
			else
				BugEater.EnableErrorDialog()
			end
		end
	)
	
	-- Show Errors in Chat
	LAM:AddCheckbox(BugEater_SettingsPanel, "BugEater_ShowErrorsInChat", "Show Errors in Chat", "Show the Errors in the Chat-Window instead", 
		function() return BugEater_SaveData.ShowErrorsInChat end,	-- getFunc
		function() 												-- setFunc
			BugEater_SaveData.ShowErrorsInChat = not BugEater_SaveData.ShowErrorsInChat
		end
	)
	
	-- Avoid Error-Flood
	LAM:AddCheckbox(BugEater_SettingsPanel, "BugEater_AvoidErrorFlood", "Avoid Error-Flood", "Avoid possible endless Error-Flood. Makes Errors Unique - if it exists it will skip buffering and output to the Chat (Exact Match).", 
		function() return BugEater_SaveData.AvoidErrorFlood end,	-- getFunc
		function() 												-- setFunc
			BugEater_SaveData.AvoidErrorFlood = not BugEater_SaveData.AvoidErrorFlood
		end
	)	

	-- Show Pre-Init Addon-Output
	LAM:AddCheckbox(BugEater_SettingsPanel, "BugEater_ShowPreInitAOP", "Show Pre-Init Addon-Output", "Show buffered Addon-Output that got thrown before the Chat was initialized completely.", 
		function() return BugEater_SaveData.ShowPreInitAOP end,	-- getFunc
		function() 													-- setFunc
			BugEater_SaveData.ShowPreInitAOP = not BugEater_SaveData.ShowPreInitAOP
			ReloadUI();
		end
	)		

	-- Suppress Addon Output in Chat
	LAM:AddCheckbox(BugEater_SettingsPanel, "BugEater_SuppressAddonOutput", "Suppress Addon-Output in Chat", "Suppress any Addon-Output via d() in the Chat.", 
		function() return BugEater_SaveData.SuppressAddonOutput end,  	-- getFunc
		function() 														-- setFunc
			BugEater_SaveData.SuppressAddonOutput = not BugEater_SaveData.SuppressAddonOutput
		end
	)

	
	-- Suppress Gui-Warnings
	LAM:AddCheckbox(BugEater_SettingsPanel, "BugEater_SuppressGuiWarnings", "Suppress All Gui-Warnings", 'Suppress all Gui-Warnings including the annoying "Too Many Anchors processed" (still shown in /error or /debug).', 
		function() return BugEater_SaveData.SuppressGuiWarnings end,  	-- getFunc
		function() 														-- setFunc
			BugEater_SaveData.SuppressGuiWarnings = not BugEater_SaveData.SuppressGuiWarnings
		end
	)	

	-- Custom Addon-Output Filter
	LAM:AddEditBox(BugEater_SettingsPanel, "BugEater_AddonOutputFilter", "Custom Addon-Output Filters", 
	'Custom Filters for suppressing Addon-Output in the Chat. Anything that is listed here will get hidden (still shown in /debug). Supports LUA-Style Regex Patterns with Partial Matching Enabled (eg. using "Loot" will match "Lootdrop" as well "Dropped Loot 123"). If you want to only match certain strings use the magic characters ^ and $.\n\nPut each Entry in a New Line.', 
		true,
		function() return BugEater_SaveData.AddonOutputFilter end,
		function()
			BugEater_SaveData.AddonOutputFilter = _G["BugEater_AddonOutputFilter"]["edit"]:GetText()
		end
	)
	
	-- Custom Error Filter
	LAM:AddEditBox(BugEater_SettingsPanel, "BugEater_ErrorFilter", "Custom Error Filters", 
	'Custom Filters for suppressing Errors. Anything that is listed here will get hidden (still shown in /error). Supports LUA-Style Regex Patterns with Partial Matching Enabled. If you want to only match certain strings use the magic characters ^ and $.\n\nPut each Entry in a New Line.', 
		true,
		function() return BugEater_SaveData.ErrorFilter end,
		function()
			BugEater_SaveData.ErrorFilter = _G["BugEater_ErrorFilter"]["edit"]:GetText()
		end
	)	

	-- Custom Error Filter

end

local function BugEater_ShowWindow(title, content)
	if(_G["BugEater_Window"] == nil) then do return false end end
	
	BugEater_Window_TITLE:SetText(title)
	BugEater_Window_EDITBUFFER:SetText(content)
	
	BugEater_Window:SetHidden(false)
	BugEater_Window:SetTopmost(true)
end

local function BugEater_HideWindow()
	if(_G["BugEater_Window"] == nil) then do return false end end
	
	BugEater_Window:SetHidden(true)
	BugEater_Window:SetTopmost(false)
end

local function BugEater_CreateWindow()
	if(_G["BugEater_Window"]) then do return false end end
	
	-- WINDOW
	local BugEater_Window = WINDOW_MANAGER:CreateTopLevelWindow("BugEater_Window")

	BugEater_Window:SetDimensions(800, 600)
	BugEater_Window:SetAnchor(CENTER, GuiRoot, CENTER)
	BugEater_Window:SetClampedToScreen(true)
	BugEater_Window:SetMouseEnabled(true)
	BugEater_Window:SetMovable(true)
	BugEater_Window:SetHidden(true)
	
	-- BG
	WINDOW_MANAGER:CreateControlFromVirtual("BugEater_Window_BG", BugEater_Window, "ZO_DefaultBackdrop")	
	
	-- CLOSE
	WINDOW_MANAGER:CreateControlFromVirtual("BugEater_Window_CLOSE", BugEater_Window, "ZO_CloseButton")
	BugEater_Window_CLOSE:SetDimensions(20, 20)
	BugEater_Window_CLOSE:SetAnchor(TOPRIGHT, BugEater_Window, TOPRIGHT, 0, 0)
	BugEater_Window_CLOSE:SetHandler("OnClicked", BugEater_HideWindow)
	
	-- TITLE
	WINDOW_MANAGER:CreateControlFromVirtual("BugEater_Window_TITLE", BugEater_Window, "ZO_WindowTitle")
	BugEater_Window_TITLE:SetAnchor(TOPLEFT, BugEater_Window, TOPLEFT, 0, -1)
	--BugEater_Window_TITLE:SetFont("ZoFontWindowTitle")
	BugEater_Window_TITLE:SetText("BugEater")
	
	-- EDITBUFFER
	BugEater_CreateEditBuffer("BugEater_Window_EDITBUFFER", BugEater_Window, {TOPLEFT, BugEater_Window_TITLE, BOTTOMLEFT, 0, 10}, {BOTTOMRIGHT, BugEater_Window, BOTTOMRIGHT, -2, -2} )
    
	-- SCROLLBAR
	--WINDOW_MANAGER:CreateControlFromVirtual("BugEater_Window_SCROLLBAR", BugEater_Window, "ZO_VerticalScrollbarBase")
	WINDOW_MANAGER:CreateControl("BugEater_Window_SCROLLBAR", BugEater_Window, CT_SLIDER)
	BugEater_Window_SCROLLBAR:SetDimensions(8, 32)
	BugEater_Window_SCROLLBAR:SetAnchor(TOPRIGHT, BugEater_Window, TOPRIGHT, -10, 60)
	BugEater_Window_SCROLLBAR:SetAnchor(BOTTOMRIGHT, BugEater_Window, BOTTOMRIGHT, -10, -40)
	BugEater_Window_SCROLLBAR:SetBackgroundMiddleTexture('EsoUI/Art/ChatWindow/chat_scrollbar_track.dds')
	BugEater_Window_SCROLLBAR:SetValueStep(1)
	BugEater_Window_SCROLLBAR:SetMouseEnabled(true)
	BugEater_Window_SCROLLBAR:SetEnabled(true)
	BugEater_Window_SCROLLBAR:SetThumbTexture('EsoUI/Art/ChatWindow/chat_thumb.dds', 'EsoUI/Art/ChatWindow/chat_thumb_disabled.dds', 'EsoUI/Art/ChatWindow/chat_thumb.dds', 8, 22, 0, 0, 0.6875)
	--BugEater_Window_SCROLLBAR:SetThumbTextureHeight(1000)
	
	BugEater_Window_SCROLLBAR:SetHandler("OnValueChanged", function(self, value, ...) 
		local max_scroll = math.floor(BugEater_Window_EDITBUFFER:GetNumCurLines()-BugEater_Window_EDITBUFFER:GetNumMaxLines())
		if(max_scroll < 0) then max_scroll = 0 end
			
		BugEater_Window_EDITBUFFER:DisplayLinesAt(value)
		BugEater_Window_SCROLLBAR:SetMinMax(0, max_scroll) 	
	end)
	
	BugEater_Window_EDITBUFFER:DisplayLinesAt(0)
	BugEater_Window_SCROLLBAR:SetValue(0)

	-- SCROLLBAR UP
	WINDOW_MANAGER:CreateControlFromVirtual("BugEater_Window_SCROLLBAR_UP", BugEater_Window_SCROLLBAR, "ZO_ScrollUpButton")
	BugEater_Window_SCROLLBAR_UP:SetAnchor(BOTTOM, BugEater_Window_SCROLLBAR, TOP, -1)
	BugEater_Window_SCROLLBAR_UP:SetNormalTexture('EsoUI/Art/ChatWindow/chat_scrollbar_upArrow_up.dds')
	BugEater_Window_SCROLLBAR_UP:SetPressedTexture('EsoUI/Art/ChatWindow/chat_scrollbar_upArrow_down.dds')
	BugEater_Window_SCROLLBAR_UP:SetMouseOverTexture('EsoUI/Art/ChatWindow/chat_scrollbar_upArrow_over.dds')
	BugEater_Window_SCROLLBAR_UP:SetDisabledTexture('EsoUI/Art/ChatWindow/chat_scrollbar_upArrow_disabled.dds')
	BugEater_Window_SCROLLBAR_UP:SetHandler("OnMouseDown", function(self, button, ...)
		BugEater_Window_EDITBUFFER:ScrollUp()
	end)

	-- SCROLLBAR DOWN    
	WINDOW_MANAGER:CreateControlFromVirtual("BugEater_Window_SCROLLBAR_DOWN", BugEater_Window_SCROLLBAR, "ZO_ScrollDownButton")
	BugEater_Window_SCROLLBAR_DOWN:SetAnchor(TOP, BugEater_Window_SCROLLBAR, BOTTOM, -1)
	BugEater_Window_SCROLLBAR_DOWN:SetNormalTexture('EsoUI/Art/ChatWindow/chat_scrollbar_downArrow_up.dds')
	BugEater_Window_SCROLLBAR_DOWN:SetPressedTexture('EsoUI/Art/ChatWindow/chat_scrollbar_downArrow_down.dds')
	BugEater_Window_SCROLLBAR_DOWN:SetMouseOverTexture('EsoUI/Art/ChatWindow/chat_scrollbar_downArrow_over.dds')
	BugEater_Window_SCROLLBAR_DOWN:SetDisabledTexture('EsoUI/Art/ChatWindow/chat_scrollbar_downArrow_disabled.dds')
	BugEater_Window_SCROLLBAR_DOWN:SetHandler("OnMouseDown", function(self, button, ...)
		BugEater_Window_EDITBUFFER:ScrollDown()
	end)    

	-- SCROLLBAR END    
	WINDOW_MANAGER:CreateControlFromVirtual("BugEater_Window_SCROLLBAR_END", BugEater_Window_SCROLLBAR, "ZO_ScrollDownButton")
	BugEater_Window_SCROLLBAR_END:SetAnchor(TOP, BugEater_Window_SCROLLBAR_DOWN, BOTTOM, 0.5)
	BugEater_Window_SCROLLBAR_END:SetNormalTexture('EsoUI/Art/ChatWindow/chat_scrollbar_endArrow_up.dds')
	BugEater_Window_SCROLLBAR_END:SetPressedTexture('EsoUI/Art/ChatWindow/chat_scrollbar_endArrow_down.dds')
	BugEater_Window_SCROLLBAR_END:SetMouseOverTexture('EsoUI/Art/ChatWindow/chat_scrollbar_endArrow_over.dds')
	BugEater_Window_SCROLLBAR_END:SetDisabledTexture('EsoUI/Art/ChatWindow/chat_scrollbar_endArrow_disabled.dds')
	BugEater_Window_SCROLLBAR_END:SetHandler("OnMouseDown", function(self, button, ...)
		BugEater_Window_EDITBUFFER:ScrollToBottom()
	end)  
        
	-- <Anchor point="TOPRIGHT" offsetX="-23" offsetY="49" />
	-- <Anchor point="BOTTOMRIGHT" offsetX="-23" offsetY="-68" />    
	
	-- COPY & PASTA Button
	WINDOW_MANAGER:CreateControlFromVirtual("BugEater_Window_COPYALL", BugEater_Window, "ZO_DefaultButton")
	BugEater_Window_COPYALL:SetDimensions(200, 30)
	BugEater_Window_COPYALL:SetAnchor(TOP, BugEater_Window, TOP, -3, 2)
	BugEater_Window_COPYALL:SetText("COPY ALL TO CLIPBOARD")
	BugEater_Window_COPYALL:SetFont("ZoFontGame")
	BugEater_Window_COPYALL:SetNormalFontColor(207/255, 220/255, 189/255, .75)
	BugEater_Window_COPYALL:SetMouseOverFontColor(255/255, 255/255, 255/255, 1)
	BugEater_Window_COPYALL:SetDisabledPressedFontColor(255/255, 255/255, 255/255, 1)
	BugEater_Window_COPYALL:SetPressedFontColor(255/255, 255/255, 255/255, 1)
	BugEater_Window_COPYALL:SetNormalTexture(nil)
	BugEater_Window_COPYALL:SetMouseOverTexture(nil)
	BugEater_Window_COPYALL:SetPressedTexture(nil)
	BugEater_Window_COPYALL:SetDisabledTexture(nil)
	BugEater_Window_COPYALL:SetMouseEnabled(true)
	BugEater_Window_COPYALL:SetEnabled(true)
	BugEater_Window_COPYALL:SetHandler("OnClicked", function(self) 
		BugEater_Window_EDITBUFFER:CopyAllTextToClipboard()
	end)
end

local function BugEater_CMD_debug()
	local content = ""
	
	for i,msg in ipairs(BugEater.Buffer["DEBUG"]) do
		content = content .. tostring(msg) .. "\n"
	end
	
	BugEater_ShowWindow("DEBUG-LOG", content)
end

local function BugEater_CMD_error()
	local content = ""
	
	for i,msg in ipairs(BugEater.Buffer["ERROR"]) do
		content = content .. tostring(msg) .. "\n"
	end

	BugEater_ShowWindow("ERROR-LOG", content)
end

local function BugEater_AddCommands()
	SLASH_COMMANDS["/debug"] = BugEater_CMD_debug
	SLASH_COMMANDS["/error"] = BugEater_CMD_error
end

local function BugEater_OnUIError(event, error_output)
	if(BugEater and BugEater_SaveData and BugEater_SaveData.AvoidErrorFlood == true and BugEater_IsErrorUnique(error_output) == false) then do return end end

	BugEater:AddMessage(error_output, "ERROR")
	
	if(BugEater and BugEater_SaveData and BugEater_SaveData.ShowErrorsInChat == true) then 			
		-- suppress gui warnings
		if(BugEater_SaveData and BugEater_SaveData.SuppressGuiWarnings == true and (error_output:find("Gui Warning:") or error_output:find("many anchors processed"))) then do return end end	
		
		-- check if any custom filter matched 
		if(BugEater_FilterMatch(error_output, "ERROR")) then do return end end
		
		if(CHAT_SYSTEM) then CHAT_SYSTEM:AddMessage("\r\n") end
		BugEater.d("[BugEater] Error caught:")
		BugEater.d(error_output) 
	end
	
	ZO_UIErrors_HideCurrent()
end

local function BugEater_OnAddonLoaded(event, addon) 
	if(addon == "BugEater") then
		BugEater_LoadSaves()
		BugEater_CreateSettings()

		if(BugEater_SaveData.SuppressErrorDialog == true) then
			BugEater.DisableErrorDialog() -- Disable the Native Debug Dialog at start
		else
			BugEater.EnableErrorDialog() -- Else Enable the Native Debug Dialog
		end

		BugEater_CreateWindow()
		BugEater_AddCommands()
	end
end 

local function BugEater_OnPlayerActivated(event) 
	
	if(BugEater and BugEater_SaveData.ShowPreInitAOP == true and ((BugEater_SaveData and (not BugEater_SaveData.SuppressAddonOutput or BugEater_SaveData.SuppressAddonOutput == false)) or not BugEater_SaveData)) then
		BugEater:PrintMessages("DEBUG")
	end
	
	if(BugEater and BugEater_SaveData and BugEater_SaveData.ShowErrorsInChat == true) then 
			for _,error_output in ipairs(BugEater.Buffer.ERROR) do
			-- suppress gui warnings
			if(BugEater_SaveData and BugEater_SaveData.SuppressGuiWarnings == true and (error_output:find("Gui Warning:") or error_output:find("many anchors processed"))) then do return end end	
			
			-- check if any custom filter matched 
			if(BugEater_FilterMatch(error_output, "ERROR")) then do return end end		
			
			if(CHAT_SYSTEM) then CHAT_SYSTEM:AddMessage("\r\n") end
			BugEater.d("[BugEater] Error caught:")
			BugEater.d(error_output) 
		end	
	end
	
	EVENT_MANAGER:UnregisterForEvent("BugEater_PLAYER_ACTIVATED", EVENT_PLAYER_ACTIVATED)
end 

-----------------------------
-- INIT
-----------------------------

function BugEater:Init()
	-- to make sure our addon is pretty much at the top of the settings
	BugEater_SettingsPanel = LAM:CreateControlPanel("BugEater_Settings", "Debug Settings")
	BugEater.DisableErrorDialog() -- temporary disable it to hinder any addons causing to throw an error before our addon has been loaded	
	
	EVENT_MANAGER:RegisterForEvent("BugEater_ADD_ON_LOADED", EVENT_ADD_ON_LOADED, BugEater_OnAddonLoaded)
	EVENT_MANAGER:RegisterForEvent("BugEater_PLAYER_ACTIVATED", EVENT_PLAYER_ACTIVATED, BugEater_OnPlayerActivated)
	EVENT_MANAGER:RegisterForEvent("BugEater_LUA_ERROR", EVENT_LUA_ERROR, BugEater_OnUIError)	
end

BugEater:Init()